import tensorflow as tf
from fm_zoo.common import EmbedFeatures, FullyConnectedNetwork, LinearModel


class CompressedInteractionNetwork(tf.keras.Model):
    """CIN part for xDeepFM.

    The structure of CIN is very similar to the RNN, where the ouputs of a hidden layer depends on the outputs from
    the previous layer(h) and an additonal input(x0).

    And the hidden layer is Conv1D that slides through the factor dimension to compress the interaction dimension.
    """
    def __init__(self, cin_hidden_sizes, split=True, name='cin'):
        super(CompressedInteractionNetwork, self).__init__(name=name)
        self.num_layers = len(cin_hidden_sizes)
        self.split = split
        self.conv_layers = [
            tf.keras.layers.Conv1D(filters=cin_hidden_size, kernel_size=1, activation='relu',
                                   data_format='channels_last', name=name + '/conv{}'.format(i))
            for i, cin_hidden_size in enumerate(cin_hidden_sizes)
        ]
        self.concat = tf.keras.layers.Concatenate(axis=1)
        self.fc = tf.keras.layers.Dense(1)

    def call(self, x, training=False):
        batch_size, factor_dim = tf.shape(x)[0], tf.shape(x)[2]
        states = []
        x0, h = tf.expand_dims(x, 2), x
        for i in range(self.num_layers):
            x = tf.multiply(x0, tf.expand_dims(h, 1))      # out product
            x = tf.reshape(x, shape=(batch_size, -1, factor_dim))
            # have to swap channel, as tf2beta don't support channels_first for Conv1D yet!
            x = tf.transpose(x, perm=[0, 2, 1])            # to bs x factor x interaction
            x = self.conv_layers[i](x)
            x = tf.transpose(x, perm=[0, 2, 1])            # to bs x compressed interaction x factor
            if self.split and i != self.num_layers - 1:
                split_size = tf.shape(x)[1] // 2
                x, h = tf.split(x, [split_size, split_size], axis=1)
            else:
                h = x
            states.append(x)
        states = self.concat(states)
        return self.fc(tf.reduce_sum(states, axis=2))


class ExtremeDeepFactorizationMachine(tf.keras.Model):
    """Implementation of xDeepFM

    reference: https://arxiv.org/abs/1803.05170
    """
    def __init__(self, feature_cards, factor_dim, fnn_hidden_sizes, cin_hidden_sizes, dropout_rate=.1, split=True,
                 name='xdeepfm'):
        super(ExtremeDeepFactorizationMachine, self).__init__(name=name)
        self.linear = LinearModel(feature_cards, name=name + '/linear_model')
        self.embedding = EmbedFeatures(feature_cards, factor_dim, name=name + '/feature_embedding')
        self.flatten = tf.keras.layers.Flatten(data_format='channels_first')
        self.nn = FullyConnectedNetwork(units=fnn_hidden_sizes, dropout_rate=dropout_rate, name=name + '/fcn')
        self.cin = CompressedInteractionNetwork(cin_hidden_sizes, split, name=name + '/cin')

    def call(self, x, training=False):
        factors = self.embedding(x)
        features = self.flatten(factors)

        linear_out = self.linear(x, training=training)
        fnn_out = self.nn(features, training=training)
        cin_out = self.cin(factors, training=training)
        return linear_out + fnn_out + cin_out



import tensorflow as tf
from fm_zoo.common import EmbedFeatures, FullyConnectedNetwork


class FMNeuralNetwork(tf.keras.Model):
    """Implementation of Factorization-machine supported Neural Networks.

    Reference: https://arxiv.org/abs/1601.02376

    It flattens feature embeddings and pass through fully connected network. Instead of prefixed inner product for
    interaction as in FM and FFM. It is asking the fcn to learn how to do feature interaction and pooling.
    """
    def __init__(self, feature_cards, factor_dim, hidden_sizes, dropout_rate=.1, name='neural_factorization_machine'):
        super(FMNeuralNetwork, self).__init__(name=name)
        self.embedding = EmbedFeatures(feature_cards, factor_dim, name=name + '/feature_embedding')
        self.flatten = tf.keras.layers.Flatten(data_format='channels_first')
        self.nn = FullyConnectedNetwork(units=hidden_sizes, dropout_rate=dropout_rate, name=name + '/fcn')

    def call(self, x, training=False):
        features = self.flatten(self.embedding(x))
        return self.nn(features, training=training)



import tensorflow as tf
from fm_zoo.common import LinearModel, FieldAwareEmbedFeatures


class FieldAwareFactorizationMachine(tf.keras.Model):
    """Implementation of Field-aware Factorization Machines.

    Reference: https://dl.acm.org/citation.cfm?id=2959134

    Extension to FM, each feature has #feautres latent factors. So when calculating pairwise interaction between
    feature_i and feature_j it is doing dot product between the jth factor of feature i and ith factor of feature j.

    The implementation here is doing element wise matrix multiplication and set lower triangle and diagnol to 0, to
    drop duplicate and self interaction before the sum pooling.
    """
    def __init__(self, feature_cards, factor_dim, name='ffm'):
        super(FieldAwareFactorizationMachine, self).__init__(name=name)
        self.factor_dim = factor_dim
        self.embeddings = FieldAwareEmbedFeatures(feature_cards, factor_dim,
                                                  name=name + '/field_aware_feature_embeddings')
        self.linear = LinearModel(feature_cards, name=name + '/linear_model')

    def call(self, x, training=False):
        batch_size, factor_dim = int(tf.shape(x)[0]), self.factor_dim
        linear_out = self.linear(x)
        factors_i = self.embeddings(x)
        factors_j = tf.transpose(factors_i, [0, 2, 1, 3])
        interactions = tf.reduce_sum(tf.multiply(factors_i, factors_j), -1)
        interaction_out = tf.expand_dims(tf.reduce_sum(
            tf.linalg.band_part(interactions, 0, -1) - tf.linalg.band_part(interactions, 0, 0),
            axis=(1, 2)
        ), axis=-1)
        return linear_out + interaction_out


import tensorflow as tf
from fm_zoo.common import EmbedFeatures, FullyConnectedNetwork, LinearModel


class DeepFM(tf.keras.Model):
    """Implementation of DeepFM

    Reference: https://arxiv.org/abs/1703.04247

    It is basically FM + FNN
    """
    def __init__(self, feature_cards, factor_dim, hidden_sizes, dropout_rate=.1, name='deepfm'):
        super(DeepFM, self).__init__(name=name)
        self.linear = LinearModel(feature_cards, name=name + '/linear_model')
        self.embedding = EmbedFeatures(feature_cards, factor_dim, name=name + '/feature_embedding')
        self.flatten = tf.keras.layers.Flatten(data_format='channels_first')
        self.nn = FullyConnectedNetwork(units=hidden_sizes, dropout_rate=dropout_rate, name=name + '/fcn')

    def call(self, x, training=False):
        factors = self.embedding(x)
        # FM part
        linear_out = self.linear(x)
        sum_of_squares = tf.reduce_sum(tf.pow(factors, 2), 1)
        square_of_sums = tf.pow(tf.reduce_sum(factors, 1), 2)
        interaction_out = .5 * tf.reduce_sum(square_of_sums - sum_of_squares, 1, keepdims=True)
        fm_out = linear_out + interaction_out

        # FNN part
        features = self.flatten(factors)
        fnn_out = self.nn(features, training=training)
        return fm_out + fnn_out



import glob
import importlib
import os.path as osp


## Import all the modules in the directory

glob_str = osp.join(osp.dirname(__file__), "*.py")
modules = glob.glob(glob_str)

# Clean up module names
modules = [ 
    osp.basename(f).replace(".py", "") 
        for f in modules if osp.isfile(f) and not f.endswith("__init__.py")
]

for module_name in modules:
    importlib.import_module(f"fm_zoo.{module_name}")


import tensorflow as tf
from fm_zoo.common import LinearModel, EmbedFeatures, FullyConnectedNetwork


class NeuralFactorizationMachine(tf.keras.Model):
    """Implementation of Neural Factorization Machines.

    Reference: https://arxiv.org/abs/1708.05027

    Contrast to FNN, the fully connected network takes in the pooled pairwise interactions, and the model keeps the
    linear part.
    """
    def __init__(self, feature_cards, factor_dim, hidden_sizes, dropout_rate=.1, name='neural_factorization_machine'):
        super(NeuralFactorizationMachine, self).__init__(name=name)
        self.embedding = EmbedFeatures(feature_cards, factor_dim, name=name + '/feature_embedding')
        self.linear = LinearModel(feature_cards, name=name + '/linear_model')
        self.nn = FullyConnectedNetwork(units=hidden_sizes, dropout_rate=dropout_rate, name=name + '/fcn')

    def call(self, x, training=False):
        linear_out = self.linear(x)
        factors = self.embedding(x)
        sum_of_squares = tf.reduce_sum(tf.pow(factors, 2), 1)
        square_of_sums = tf.pow(tf.reduce_sum(factors, 1), 2)
        pooled_interactions = square_of_sums - sum_of_squares
        interaction_out = self.nn(pooled_interactions, training=training)
        return linear_out + interaction_out



import tensorflow as tf
from fm_zoo.common import LinearModel, EmbedFeatures


class FactorizationMachine(tf.keras.Model):
    """Implementation of Factorization Machines.

    Reference: https://www.csie.ntu.edu.tw/~b97053/paper/Rendle2010FM.pdf

    fm(x) = linear regression + pooling of degree d interactions between latent factors of dimension k.

    Degree d here is fixed at 2, so that we can simplify the calculation. Only hyperparam is factor_dim

    The input is discretized/Integer Encoded instead of real valued.
    See common.EmbedFeatures for detail on input configs.
   """
    def __init__(self, feature_cards, factor_dim, name='factorization_machine'):
        super(FactorizationMachine, self).__init__(name=name)
        self.embedding = EmbedFeatures(feature_cards, factor_dim, name=name + '/feature_embedding')
        self.linear = LinearModel(feature_cards, name=name + '/linear_model')

    def call(self, x, training=False):
        linear_out = self.linear(x)
        factors = self.embedding(x)
        sum_of_squares = tf.reduce_sum(tf.pow(factors, 2), 1)
        square_of_sums = tf.pow(tf.reduce_sum(factors, 1), 2)
        interaction_out = 0.5 * tf.reduce_sum(square_of_sums - sum_of_squares, 1, keepdims=True)
        return linear_out + interaction_out


import os.path as osp
from setuptools import setup, find_packages


# read the contents of your README file
rep_folder = osp.abspath(osp.dirname(__file__))
with open(osp.join(rep_folder, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="tf2_fm_zoo",
    version="0.1.0",
    url="https://github.com/RoetGer/tf2-fm-zoo",
    packages=find_packages(include=["fm_zoo"]),
    long_description=long_description, 
    long_description_content_type="text/markdown"
)


import numpy as np
import tensorflow as tf


class LinearModel(tf.keras.Model):
    """The linear model.

    The input to this module is the same as the EmbedFeatures module, see that part for detail.
    """
    def __init__(self, feature_cards, name='linear_model'):
        super(LinearModel, self).__init__(name=name)
        self.bias = tf.random.uniform((1,))
        self.linear = tf.keras.layers.Embedding(input_dim=sum(feature_cards), output_dim=1)
        self.offsets = tf.constant(np.concatenate(([0], np.cumsum(feature_cards)[:-1])), dtype='int64')

    def call(self, x, training=False):
        x = x + tf.stop_gradient(self.offsets)
        return self.bias + tf.reduce_sum(self.linear(x), axis=1)


class EmbedFeatures(tf.keras.Model):
    """Embed encoded features.

    Each input to the embedding module is encoded values of its features.
    Each feature is indexed from 0.
    For continuous features, we would need to bin and encode first.

    For example, if our dataset has three features, value and (encoding):
        feature 1: gender - F(0), M(1), O(2)
        feature 2: edu - highschool(0), bachelor(1), master(2), phd(3)
        feature 3: age - 18-(0), 18-24(1), 25-40(2), 40-65(3), 65+(4)
    and we want to embed features to dimension k=2.

    To initialize the embedding, we would pass feature_cards = [3, 4, 5] and factor_dim = 2

    Example of encoded inputs:
        (male, master, 26) encoded as [1, 2, 2]
        (female, phd, 35) encoded as [0, 3, 2]
    """
    def __init__(self, feature_cards, factor_dim, name='embedding'):
        super(EmbedFeatures, self).__init__(name=name)
        self.embedding = tf.keras.layers.Embedding(input_dim=sum(feature_cards), output_dim=factor_dim)
        self.offsets = tf.constant(np.concatenate(([0], np.cumsum(feature_cards)[:-1])), dtype='int64')

    def call(self, x, training=False):
        x = x + tf.stop_gradient(self.offsets)
        embedded = self.embedding(x)
        return embedded


class FieldAwareEmbedFeatures(tf.keras.Model):
    """Embed encoded features to num_features different latent factors.

    To get the ith feature's jth factor of example k in the batch, we would do: embedded[k, i, j, :]
    """
    def __init__(self, feature_cards, factor_dim, name='embedding'):
        super(FieldAwareEmbedFeatures, self).__init__(name=name)
        self.num_features = len(feature_cards)
        self.embeddings = [
            tf.keras.layers.Embedding(sum(feature_cards), factor_dim, name=name + '/embed_{}'.format(i))
            for i in range(self.num_features)
        ]
        self.offsets = tf.constant(np.concatenate(([0], np.cumsum(feature_cards)[:-1])), dtype='int64')

    def call(self, x, training=False):
        x = x + tf.stop_gradient(self.offsets)
        embeddings = [tf.expand_dims(embedding(x), 2) for embedding in self.embeddings]
        return tf.concat(embeddings, 2)


class FullyConnectedNetwork(tf.keras.Model):
    """Fully connected feedforward network with optional dropouts."""
    def __init__(self, units, dropout_rate=.1, name='fcn'):
        super(FullyConnectedNetwork, self).__init__(name=name)
        layers = []
        for i, unit in enumerate(units):
            layers.append(tf.keras.layers.Dense(units=unit, name=name + '/fc{}'.format(i)))
            if dropout_rate > 0:
                layers.append(tf.keras.layers.Dropout(rate=dropout_rate, name=name + '/dropout{}'.format(i)))
            if unit != 1:
                layers.append(tf.keras.layers.ReLU(name=name + '/activ{}'.format(i)))
        self.model = tf.keras.Sequential(layers)

    def call(self, x, training=False):
        return self.model(x, training=training)


def attention(value, key, query, mask=None, dropout=None, training=False):
    depth = query.shape[-1]
    logits = tf.matmul(query, key, transpose_b=True) / depth ** .5
    if mask:
        logits += mask * -1e9
    if dropout:
        logits = dropout(logits, training=training)
    alignment = tf.nn.softmax(logits, axis=-1)
    attended = tf.matmul(alignment, value)
    return attended


class MultiHeadAttention(tf.keras.Model):
    """Multi-Head Attention"""
    def __init__(self, n_heads, d_model, dropout_rate=.1, name='multihead_attention'):
        super(MultiHeadAttention, self).__init__(name=name)
        assert d_model % n_heads == 0
        self.n_heads = n_heads
        self.d_model = d_model
        self.depth = d_model // n_heads
        self.query_linear = tf.keras.layers.Dense(d_model)
        self.key_linear = tf.keras.layers.Dense(d_model)
        self.value_linear = tf.keras.layers.Dense(d_model)
        self.out_linear = tf.keras.layers.Dense(d_model)
        self.dropout = tf.keras.layers.Dropout(rate=dropout_rate)

    def call(self, value, key, query, mask=None, training=False):
        bs = tf.shape(query)[0]
        query = self.query_linear(query)
        key = self.key_linear(key)
        value = self.value_linear(value)

        query = tf.transpose(tf.reshape(query, (bs, -1, self.n_heads, self.depth)), [0, 2, 1, 3])
        key = tf.transpose(tf.reshape(key, (bs, -1, self.n_heads, self.depth)), [0, 2, 1, 3])
        value = tf.transpose(tf.reshape(value, (bs, -1, self.n_heads, self.depth)), [0, 2, 1, 3])

        attn = attention(query, key, value, mask=mask, dropout=self.dropout, training=training)
        attn = tf.transpose(attn, [0, 2, 1, 3])
        attn = tf.reshape(attn, (bs, -1, self.d_model))

        out = self.out_linear(attn)
        return out



import tensorflow as tf
from fm_zoo.common import EmbedFeatures, FullyConnectedNetwork, LinearModel, MultiHeadAttention


class AutomaticFeatureInteraction(tf.keras.Model):
    """Implementation of AutoInt

    Reference: https://arxiv.org/abs/1810.11921

    Difference with FNN is that AutoInt using MultiHeadAttention to model feature interactions.
    """
    def __init__(self, feature_cards, factor_dim, n_heads, n_attentions, hidden_sizes, dropout_rate=.1, name='deepfm'):
        super(AutomaticFeatureInteraction, self).__init__(name=name)
        self.linear = LinearModel(feature_cards, name=name + '/linear_model')
        self.embedding = EmbedFeatures(feature_cards, factor_dim, name=name + '/feature_embedding')
        self.flatten = tf.keras.layers.Flatten(data_format='channels_first')
        self.nn = FullyConnectedNetwork(units=hidden_sizes, dropout_rate=dropout_rate, name=name + '/fcn')
        self.attns = [
            MultiHeadAttention(n_heads, factor_dim, dropout_rate=dropout_rate, name=name + '/mhattn{}'.format(i))
            for i in range(n_attentions)
        ]
        self.attn_out = tf.keras.layers.Dense(1)

    def call(self, x, training=False):
        factors = self.embedding(x)                     # bs x num_features x factor_dim

        # linear part
        linear_out = self.linear(x)                     # bs x 1

        # FNN part
        features = self.flatten(factors)                # bs x num_features * factor_dim
        fnn_out = self.nn(features)                     # bs x 1

        # ATTN part
        cross_term = tf.transpose(factors, [0, 2, 1])   # bs x factor_dim x num_features
        for attn in self.attns:
            cross_term = attn(cross_term, cross_term, cross_term, training=training)
        cross_term = tf.transpose(cross_term, [0, 2, 1])
        cross_term = self.flatten(tf.nn.relu(cross_term))
        attn_out = self.attn_out(cross_term)
        return linear_out + fnn_out + attn_out



import tensorflow as tf
from fm_zoo.common import LinearModel, EmbedFeatures


class AttentionalFactorizationMachine(tf.keras.Model):
    """Implementation of Attentional Factorization Machines.

    Reference: https://arxiv.org/abs/1708.04617

    It adds an attention network on the pairwise interactions of FM.
    """
    def __init__(self, feature_cards, factor_dim, attention_size, name='afm'):
        super(AttentionalFactorizationMachine, self).__init__(name=name)
        self.num_features = len(feature_cards)
        self.factor_dim = factor_dim
        self.linear = LinearModel(feature_cards, name=name + '/linear_model')
        self.embedding = EmbedFeatures(feature_cards, factor_dim, name=name + '/feature_embedding')
        self.attention = tf.keras.Sequential([
            tf.keras.layers.Dense(units=attention_size, name=name + '/attention_hidden'),
            tf.keras.layers.ReLU(name=name + '/attention_activ'),
            tf.keras.layers.Dense(units=1, name=name + '/attention_logits'),
            tf.keras.layers.Softmax(axis=1, name=name + '/attention_score')
        ])

    def call(self, x, training=False):
        batch_size, num_features = int(tf.shape(x)[0]), self.num_features
        num_interactions = num_features * (num_features - 1) // 2

        factors = self.embedding(x)
        factors_i = tf.tile(tf.expand_dims(factors, 1), [1, num_features, 1, 1])
        factors_j = tf.tile(tf.expand_dims(factors, 2), [1, 1, num_features, 1])
        interactions = tf.multiply(factors_i, factors_j)

        # use a mask to get unique pairwise interactions
        mask = tf.ones(interactions.shape[:-1])
        mask = tf.cast(tf.linalg.band_part(mask, 0, -1) - tf.linalg.band_part(mask, 0, 0), dtype=tf.bool)
        interactions = tf.boolean_mask(interactions, mask)

        interactions = tf.reshape(interactions, [batch_size, num_interactions, self.factor_dim])
        attention_scores = self.attention(interactions)
        interactions = tf.reduce_sum(interactions, axis=-1, keepdims=True)
        attended = tf.multiply(interactions, attention_scores)
        return self.linear(x) + tf.reduce_sum(attended, axis=1)



import tensorflow as tf
from fm_zoo.common import LinearModel, FieldAwareEmbedFeatures, FullyConnectedNetwork


class FieldAwareNeuralFactorizationMachine(tf.keras.Model):
    """Implementation of Field-aware Neural Factorization Machines.

    Reference: https://arxiv.org/abs/1902.09096

    extend nfm to operate on field aware interactions
    """
    def __init__(self, feature_cards, factor_dim, hidden_sizes, dropout_rate=.1, name='fnfm'):
        super(FieldAwareNeuralFactorizationMachine, self).__init__(name=name)
        self.num_features = len(feature_cards)
        self.embeddings = FieldAwareEmbedFeatures(feature_cards, factor_dim,
                                                  name=name + '/field_aware_feature_embedding')
        self.linear = LinearModel(feature_cards, name=name + '/linear_model')
        self.nn = FullyConnectedNetwork(units=hidden_sizes, dropout_rate=dropout_rate, name=name + '/fcn')

    def call(self, x, training=False):
        batch_size, num_features = int(tf.shape(x)[0]), self.num_features
        num_interactions = num_features * (num_features - 1) // 2

        linear_out = self.linear(x)

        factors_i = self.embeddings(x)
        factors_j = tf.transpose(factors_i, [0, 2, 1, 3])
        interactions = tf.reduce_sum(tf.multiply(factors_i, factors_j), axis=-1)
        mask = tf.ones_like(interactions)
        mask = tf.cast(tf.linalg.band_part(mask, 0, -1) - tf.linalg.band_part(mask, 0, 0), dtype=tf.bool)
        interactions = tf.boolean_mask(interactions, mask)
        interactions = tf.reshape(interactions, shape=[batch_size, num_interactions])
        interaction_out = self.nn(interactions, training=training)

        return linear_out + interaction_out
